from bbot.modules.report.base import BaseReportModule


class aggregate(BaseReportModule):
    flags = ["passive", "safe"]
    meta = {"description": "Summarize statistics at the end of a scan"}

    async def report(self):
        self.log_table(*self.scan.stats._make_table(), table_name="scan-stats")



import re

from .emailformat import emailformat


class skymem(emailformat):
    watched_events = ["DNS_NAME"]
    produced_events = ["EMAIL_ADDRESS"]
    flags = ["passive", "email-enum", "safe"]
    meta = {"description": "Query skymem.info for email addresses"}

    base_url = "https://www.skymem.info"

    async def handle_event(self, event):
        _, query = self.helpers.split_domain(event.data)
        # get first page
        url = f"{self.base_url}/srch?q={self.helpers.quote(query)}"
        r = await self.request_with_fail_count(url)
        if not r:
            return
        for email in self.helpers.extract_emails(r.text):
            await self.emit_event(email, "EMAIL_ADDRESS", source=event)

        # iterate through other pages
        domain_ids = re.findall(r'<a href="/domain/([a-z0-9]+)\?p=', r.text, re.I)
        if not domain_ids:
            return
        domain_id = domain_ids[0]
        for page in range(2, 22):
            r2 = await self.request_with_fail_count(f"{self.base_url}/domain/{domain_id}?p={page}")
            if not r2:
                continue
            for email in self.helpers.extract_emails(r2.text):
                await self.emit_event(email, "EMAIL_ADDRESS", source=event)
            pages = re.findall(r"/domain/" + domain_id + r"\?p=(\d+)", r2.text)
            if not pages:
                break
            last_page = max([int(p) for p in pages])
            if page >= last_page:
                break



import re
from werkzeug.wrappers import Response

from .base import ModuleTestBase


def extract_subdomain_tag(data):
    pattern = r"http://([a-z0-9]{4})\.fakedomain\.fakeinteractsh\.com"
    match = re.search(pattern, data)
    if match:
        return match.group(1)


class TestGeneric_SSRF(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "generic_ssrf"]

    def request_handler(self, request):
        subdomain_tag = None

        if request.method == "GET":
            subdomain_tag = extract_subdomain_tag(request.full_path)
        elif request.method == "POST":
            subdomain_tag = extract_subdomain_tag(request.data.decode())
        if subdomain_tag:
            self.interactsh_mock_instance.mock_interaction(subdomain_tag)

        return Response("alive", status=200)

    async def setup_before_prep(self, module_test):
        self.interactsh_mock_instance = module_test.request_fixture.getfixturevalue("interactsh_mock_instance")
        module_test.monkeypatch.setattr(
            module_test.scan.helpers, "interactsh", lambda *args, **kwargs: self.interactsh_mock_instance
        )

    async def setup_after_prep(self, module_test):
        expect_args = re.compile("/")
        module_test.set_expect_requests_handler(expect_args=expect_args, request_handler=self.request_handler)

    def check(self, module_test, events):
        assert any(
            e.type == "VULNERABILITY" and "Out-of-band interaction: [Generic SSRF (GET)]" in e.data["description"]
            for e in events
        ), "Failed to detect Generic SSRF (GET)"
        assert any(
            e.type == "VULNERABILITY" and "Out-of-band interaction: [Generic SSRF (POST)]" in e.data["description"]
            for e in events
        ), "Failed to detect Generic SSRF (POST)"
        assert any(
            e.type == "VULNERABILITY" and "Out-of-band interaction: [Generic XXE] [HTTP]" in e.data["description"]
            for e in events
        ), "Failed to detect Generic SSRF (XXE)"



from contextlib import suppress

from bbot.modules.base import BaseModule


class credshed(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["PASSWORD", "HASHED_PASSWORD", "USERNAME", "EMAIL_ADDRESS"]
    flags = ["passive", "safe"]
    meta = {
        "description": "Send queries to your own credshed server to check for known credentials of your targets",
        "auth_required": True,
    }
    options = {"username": "", "password": "", "credshed_url": ""}
    options_desc = {
        "username": "Credshed username",
        "password": "Credshed password",
        "credshed_url": "URL of credshed server",
    }
    target_only = True

    async def setup(self):
        self.base_url = self.config.get("credshed_url", "").rstrip("/")
        self.username = self.config.get("username", "")
        self.password = self.config.get("password", "")

        # soft-fail if we don't have the necessary information to make queries
        if not (self.base_url and self.username and self.password):
            return None, "Must set username, password, and credshed_url"

        auth_setup = await self.helpers.request(
            f"{self.base_url}/api/auth", method="POST", json={"username": self.username, "password": self.password}
        )
        self.auth_token = ""
        with suppress(Exception):
            self.auth_token = auth_setup.json().get("access_token", "")
        # hard-fail if we didn't get an access token
        if not self.auth_token:
            return False, f"Failed to retrieve credshed auth token from url: {self.base_url}"

        return await super().setup()

    async def handle_event(self, event):
        query = event.data
        cs_query = await self.helpers.request(
            f"{self.base_url}/api/search",
            method="POST",
            cookies={"access_token_cookie": self.auth_token},
            json={"query": query},
        )

        if cs_query is not None and cs_query.status_code != 200:
            self.warning(
                f"Error retrieving results from {self.base_url} (status code {cs_query.status_code}): {cs_query.text}"
            )

        json_result = {}
        with suppress(Exception):
            json_result = cs_query.json()

        if not json_result:
            return

        accounts = json_result.get("accounts", [])

        for i in accounts:
            email = i.get("e", "")
            pw = i.get("p", "")
            hashes = i.get("h", [])
            user = i.get("u", "")
            src = i.get("s", [])
            src = [src[0] if src else ""]

            tags = []
            if src:
                tags = [f"credshed-source-{src}"]

            email_event = self.make_event(email, "EMAIL_ADDRESS", source=event, tags=tags)
            if email_event is not None:
                await self.emit_event(email_event)
                if user:
                    await self.emit_event(f"{email}:{user}", "USERNAME", source=email_event, tags=tags)
                if pw:
                    await self.emit_event(f"{email}:{pw}", "PASSWORD", source=email_event, tags=tags)
                for h_pw in hashes:
                    if h_pw:
                        await self.emit_event(f"{email}:{h_pw}", "HASHED_PASSWORD", source=email_event, tags=tags)



import os
import logging
import threading
import traceback
from contextlib import suppress

from .misc import rm_at_exit


log = logging.getLogger("bbot.core.helpers.files")


def tempfile(self, content, pipe=True):
    """
    Creates a temporary file or named pipe and populates it with content.

    Args:
        content (list, set, tuple, str): The content to populate the temporary file with.
        pipe (bool, optional): If True, a named pipe is used instead of a true file.
            This allows Python data to be piped directly into the process without taking up disk space.
            Defaults to True.

    Returns:
        str: The filepath of the created temporary file or named pipe.

    Examples:
        >>> tempfile(["This", "is", "temp", "content"])
        '/home/user/.bbot/temp/pgxml13bov87oqrvjz7a'

        >>> tempfile(["Another", "temp", "file"], pipe=False)
        '/home/user/.bbot/temp/someotherfile'
    """
    filename = self.temp_filename()
    rm_at_exit(filename)
    try:
        if type(content) not in (set, list, tuple):
            content = (content,)
        if pipe:
            os.mkfifo(filename)
            self.feed_pipe(filename, content, text=True)
        else:
            with open(filename, "w", errors="ignore") as f:
                for c in content:
                    line = f"{self.smart_decode(c)}\n"
                    f.write(line)
    except Exception as e:
        log.error(f"Error creating temp file: {e}")
        log.trace(traceback.format_exc())

    return filename


def _feed_pipe(self, pipe, content, text=True):
    """
    Feeds content into a named pipe or file-like object.

    Args:
        pipe (str or file-like object): The named pipe or file-like object to feed the content into.
        content (iterable): The content to be written into the pipe or file.
        text (bool, optional): If True, the content is decoded using smart_decode function.
            If False, smart_encode function is used. Defaults to True.

    Notes:
        The method tries to determine if 'pipe' is a file-like object that has a 'write' method.
        If so, it writes directly to that object. Otherwise, it opens 'pipe' as a file for writing.
    """
    try:
        if text:
            decode_fn = self.smart_decode
            newline = "\n"
        else:
            decode_fn = self.smart_encode
            newline = b"\n"
        try:
            if hasattr(pipe, "write"):
                try:
                    for c in content:
                        pipe.write(decode_fn(c) + newline)
                finally:
                    with suppress(Exception):
                        pipe.close()
            else:
                with open(pipe, "w") as p:
                    for c in content:
                        p.write(decode_fn(c) + newline)
        except BrokenPipeError:
            log.debug(f"Broken pipe in _feed_pipe()")
        except ValueError:
            log.debug(f"Error _feed_pipe(): {traceback.format_exc()}")
    except KeyboardInterrupt:
        self.scan.stop()
    except Exception as e:
        log.error(f"Error in _feed_pipe(): {e}")
        log.trace(traceback.format_exc())


def feed_pipe(self, pipe, content, text=True):
    """
    Starts a new thread to feed content into a named pipe or file-like object using _feed_pipe().

    Args:
        pipe (str or file-like object): The named pipe or file-like object to feed the content into.
        content (iterable): The content to be written into the pipe or file.
        text (bool, optional): If True, the content is decoded using smart_decode function.
            If False, smart_encode function is used. Defaults to True.
    """
    t = threading.Thread(target=self._feed_pipe, args=(pipe, content), kwargs={"text": text}, daemon=True)
    t.start()


def tempfile_tail(self, callback):
    """
    Create a named pipe and execute a callback function on each line that is written to the pipe.

    Useful for ingesting output from a program (e.g. nuclei) directly from a file in real-time as
    each line is written. The idea is you create the file with this function and then tell the CLI
    program to output to it as a normal output file. We are then able to scoop up the output line
    by line as it's written to our "file" (which is actually a named pipe, shhh! ;)

    Args:
        callback (Callable): A function that will be invoked with each line written to the pipe as its argument.

    Returns:
        str: The filename of the created named pipe.
    """
    filename = self.temp_filename()
    rm_at_exit(filename)
    try:
        os.mkfifo(filename)
        t = threading.Thread(target=tail, args=(filename, callback), daemon=True)
        t.start()
    except Exception as e:
        log.error(f"Error setting up tail for file {filename}: {e}")
        log.trace(traceback.format_exc())
        return
    return filename


def tail(filename, callback):
    """
    Continuously read lines from a file and execute a callback function on each line.

    Args:
        filename (str): The path of the file to tail.
        callback (Callable): A function to call on each line read from the file.

    Examples:
        >>> def print_callback(line):
        ...     print(f"Received: {line}")
        >>> tail("/path/to/file", print_callback)
    """
    try:
        with open(filename, errors="ignore") as f:
            for line in f:
                line = line.rstrip("\r\n")
                callback(line)
    except Exception as e:
        log.error(f"Error tailing file {filename}: {e}")
        log.trace(traceback.format_exc())



import asyncio
import logging

from cloudcheck import cloud_providers

log = logging.getLogger("bbot.helpers.cloud")


class CloudHelper:
    def __init__(self, parent_helper):
        self.parent_helper = parent_helper
        self.providers = cloud_providers
        self.dummy_modules = {}
        for provider_name in self.providers.providers:
            self.dummy_modules[provider_name] = self.parent_helper._make_dummy_module(
                f"{provider_name}_cloud", _type="scan"
            )
        self._updated = False
        self._update_lock = asyncio.Lock()

    def excavate(self, event, s):
        """
        Extract buckets, etc. from strings such as an HTTP responses
        """
        for provider in self:
            provider_name = provider.name.lower()
            base_kwargs = {"source": event, "tags": [f"cloud-{provider_name}"], "_provider": provider_name}
            for event_type, sigs in provider.signatures.items():
                found = set()
                for sig in sigs:
                    for match in sig.findall(s):
                        kwargs = dict(base_kwargs)
                        kwargs["event_type"] = event_type
                        if not match in found:
                            found.add(match)
                            if event_type == "STORAGE_BUCKET":
                                self.emit_bucket(match, **kwargs)
                            else:
                                self.emit_event(**kwargs)

    def speculate(self, event):
        """
        Look for DNS_NAMEs that are buckets or other cloud resources
        """
        for provider in self:
            provider_name = provider.name.lower()
            base_kwargs = dict(
                source=event, tags=[f"{provider.provider_type}-{provider_name}"], _provider=provider_name
            )
            if event.type.startswith("DNS_NAME"):
                for event_type, sigs in provider.signatures.items():
                    found = set()
                    for sig in sigs:
                        match = sig.match(event.data)
                        if match:
                            kwargs = dict(base_kwargs)
                            kwargs["event_type"] = event_type
                            if not event.data in found:
                                found.add(event.data)
                                if event_type == "STORAGE_BUCKET":
                                    self.emit_bucket(match.groups(), **kwargs)
                                else:
                                    self.emit_event(**kwargs)

    def emit_bucket(self, match, **kwargs):
        bucket_name, bucket_domain = match
        kwargs["data"] = {"name": bucket_name, "url": f"https://{bucket_name}.{bucket_domain}"}
        self.emit_event(**kwargs)

    def emit_event(self, *args, **kwargs):
        provider_name = kwargs.pop("_provider")
        dummy_module = self.dummy_modules[provider_name]
        event = dummy_module.make_event(*args, **kwargs)
        if event:
            self.parent_helper.scan.manager.queue_event(event)

    async def tag_event(self, event):
        """
        Tags an event according to cloud provider
        """
        async with self._update_lock:
            if not self._updated:
                await self.providers.update()
                self._updated = True

        if event.host:
            for host in [event.host] + list(event.resolved_hosts):
                provider_name, provider_type, source = self.providers.check(host)
                if provider_name is not None:
                    provider = self.providers.providers[provider_name.lower()]
                    event.add_tag(f"{provider_type}-{provider_name.lower()}")
                    # if its host directly matches this cloud provider's domains
                    if not self.parent_helper.is_ip(host):
                        # tag as buckets, etc.
                        for event_type, sigs in provider.signatures.items():
                            for sig in sigs:
                                if sig.match(host):
                                    event.add_tag(f"{provider_type}-{event_type}")

    def __getitem__(self, item):
        return self.providers.providers[item.lower()]

    def __iter__(self):
        yield from self.providers



from .base import ModuleTestBase


class TestMasscan(ModuleTestBase):
    targets = ["8.8.8.8/32"]
    scan_name = "test_masscan"
    config_overrides = {"modules": {"masscan": {"ports": "443", "wait": 1}}}

    masscan_output = """[
{   "ip": "8.8.8.8",   "timestamp": "1680197558", "ports": [ {"port": 443, "proto": "tcp", "status": "open", "reason": "syn-ack", "ttl": 54} ] }
]"""

    async def setup_after_prep(self, module_test):
        self.masscan_run = False

        async def run_masscan(command, *args, **kwargs):
            if "masscan" in command[:2]:
                for l in self.masscan_output.splitlines():
                    yield l
                self.masscan_run = True
            else:
                async for l in module_test.scan.helpers.run_live(command, *args, **kwargs):
                    yield l

        module_test.monkeypatch.setattr(module_test.scan.helpers, "run_live", run_masscan)

    def check(self, module_test, events):
        assert self.masscan_run == True
        assert any(e.type == "IP_ADDRESS" and e.data == "8.8.8.8" for e in events), "No IP_ADDRESS emitted"
        assert any(e.type == "OPEN_TCP_PORT" and e.data == "8.8.8.8:443" for e in events), "No OPEN_TCP_PORT emitted"


class TestMasscan1(TestMasscan):
    modules_overrides = ["masscan"]
    config_overrides = {"modules": {"masscan": {"ports": "443", "wait": 1, "use_cache": True}}}

    def check(self, module_test, events):
        assert self.masscan_run == False
        assert any(e.type == "IP_ADDRESS" and e.data == "8.8.8.8" for e in events), "No IP_ADDRESS emitted"
        assert any(e.type == "OPEN_TCP_PORT" and e.data == "8.8.8.8:443" for e in events), "No OPEN_TCP_PORT emitted"



from .base import ModuleTestBase, tempwordlist


class TestMassdns(ModuleTestBase):
    subdomain_wordlist = tempwordlist(["www", "asdf"])
    config_overrides = {"modules": {"massdns": {"wordlist": str(subdomain_wordlist)}}}

    def check(self, module_test, events):
        assert any(e.data == "www.blacklanternsecurity.com" for e in events)
        assert not any(e.data == "asdf.blacklanternsecurity.com" for e in events)



from .base import ModuleTestBase


class TestDigitorus(ModuleTestBase):
    web_response = """<a href="/b8198b95b449ef633d3b671fdd5e5096a81bbc161afb07fa50d29edaac33bf88/asdf.blacklanternsecurity.com" title="Show the certificate for www.blacklanternsecurity.com">www.blacklanternsecurity.com</a><br>
<a href="/d92f154de36b1c3ea253a60a41c1a30e148e8964f92e10df4789692860ea80cb/zzzz.blacklanternsecurity.com" title="Show the certificate for chat.blacklanternsecurity.com">chat.blacklanternsecurity.com</a><br>
<a href="/e8b44651bd01af5d077045c2792c6038f0bf3d26684bf2170546d9affed4bf52/zzzz.blacklanternsecurity.com" title="Show the certificate for www.blacklanternsecurity.com">www.blacklanternsecurity.com</a><br>
<a href="/faef21c8c799d9ee1867ab6028ff33ade4d03c39277e65c9abe23e3633a10496/asdf.blacklanternsecurity.com" title="Show the certificate for tasks.blacklanternsecurity.com">tasks.blacklanternsecurity.com</a><br>
<a href="/ff1075573cc59a60073e968e61728a30b66974c234a9feeb07d695dfd3391512/asdf.blacklanternsecurity.com" title="Show the certificate for gitlab.blacklanternsecurity.com">gitlab.blacklanternsecurity.com</a><br>
"""

    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url=f"https://certificatedetails.com/blacklanternsecurity.com",
            text=self.web_response,
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"
        assert any(e.data == "zzzz.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class securitytrails(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {"description": "Query the SecurityTrails API for subdomains", "auth_required": True}
    options = {"api_key": ""}
    options_desc = {"api_key": "SecurityTrails API key"}

    base_url = "https://api.securitytrails.com/v1"

    async def setup(self):
        self.limit = 100
        return await super().setup()

    async def ping(self):
        url = f"{self.base_url}/ping?apikey={self.api_key}"
        r = await self.request_with_fail_count(url)
        resp_content = getattr(r, "text", "")
        assert getattr(r, "status_code", 0) == 200, resp_content

    async def request_url(self, query):
        url = f"{self.base_url}/domain/{query}/subdomains?apikey={self.api_key}"
        response = await self.request_with_fail_count(url)
        return response

    def parse_results(self, r, query):
        j = r.json()
        if isinstance(j, dict):
            for host in j.get("subdomains", []):
                yield f"{host}.{query}"



from bbot.modules.base import BaseModule
from wafw00f import main as wafw00f_main

# disable wafw00f logging
import logging

wafw00f_logger = logging.getLogger("wafw00f")
wafw00f_logger.setLevel(logging.CRITICAL + 100)


class wafw00f(BaseModule):
    """
    https://github.com/EnableSecurity/wafw00f
    """

    watched_events = ["URL"]
    produced_events = ["WAF"]
    flags = ["active", "aggressive"]
    meta = {"description": "Web Application Firewall Fingerprinting Tool"}

    deps_pip = ["wafw00f~=2.2.0"]

    options = {"generic_detect": True}
    options_desc = {"generic_detect": "When no specific WAF detections are made, try to perform a generic detect"}

    in_scope_only = True
    per_hostport_only = True

    async def filter_event(self, event):
        http_status = getattr(event, "http_status", 0)
        if not http_status or str(http_status).startswith("3"):
            return False, f"Invalid HTTP status code: {http_status}"
        return True, ""

    async def handle_event(self, event):
        url = f"{event.parsed.scheme}://{event.parsed.netloc}/"
        WW = await self.scan.run_in_executor(wafw00f_main.WAFW00F, url, followredirect=False)
        waf_detections = await self.scan.run_in_executor(WW.identwaf)
        if waf_detections:
            for waf in waf_detections:
                await self.emit_event({"host": str(event.host), "url": url, "WAF": waf}, "WAF", source=event)
        else:
            if self.config.get("generic_detect") == True:
                generic = await self.scan.run_in_executor(WW.genericdetect)
                if generic:
                    await self.emit_event(
                        {
                            "host": str(event.host),
                            "url": url,
                            "WAF": "generic detection",
                            "info": WW.knowledge["generic"]["reason"],
                        },
                        "WAF",
                        source=event,
                    )



from ..bbot_fixtures import *  # noqa: F401


@pytest.mark.asyncio
async def test_config(bbot_config, bbot_scanner):
    scan1 = bbot_scanner("127.0.0.1", modules=["ipneighbor", "speculate"], config=bbot_config)
    await scan1.load_modules()
    assert scan1.config.plumbus == "asdf"
    assert scan1.modules["ipneighbor"].config.test_option == "ipneighbor"
    assert scan1.modules["python"].config.test_option == "asdf"
    assert scan1.modules["speculate"].config.test_option == "speculate"



from bbot.modules.base import BaseModule
import xml.etree.ElementTree as ET


class bucket_file_enum(BaseModule):
    """
    Enumerate files in a public bucket
    """

    watched_events = ["STORAGE_BUCKET"]
    produced_events = ["URL_UNVERIFIED"]
    meta = {
        "description": "Works in conjunction with the filedownload module to download files from open storage buckets. Currently supported cloud providers: AWS"
    }
    flags = ["passive", "safe", "cloud-enum"]
    options = {
        "file_limit": 50,
    }
    options_desc = {"file_limit": "Limit the number of files downloaded per bucket"}
    scope_distance_modifier = 2

    async def setup(self):
        self.file_limit = self.config.get("file_limit", 50)
        return True

    async def handle_event(self, event):
        cloud_tags = (t for t in event.tags if t.startswith("cloud-"))
        if any(t.endswith("-amazon") or t.endswith("-digitalocean") for t in cloud_tags):
            await self.handle_aws(event)

    async def handle_aws(self, event):
        url = event.data["url"]
        urls_emitted = 0
        response = await self.helpers.request(url)
        status_code = getattr(response, "status_code", 0)
        if status_code == 200:
            content = response.text
            root = ET.fromstring(content)
            namespace = {"s3": "http://s3.amazonaws.com/doc/2006-03-01/"}
            keys = [key.text for key in root.findall(".//s3:Key", namespace)]
            for key in keys:
                bucket_file = url + "/" + key
                file_extension = self.helpers.get_file_extension(key)
                if file_extension not in self.scan.url_extension_blacklist:
                    await self.emit_event(bucket_file, "URL_UNVERIFIED", source=event, tags="filedownload")
                    urls_emitted += 1
                    if urls_emitted >= self.file_limit:
                        return
